/*
 * Copyright (c) 1997-2006 Erez Zadok,
 * Copyright (c) 2004-2006 Nikolai Joukov,
 * Copyright (c) 2005      Arun M. Krishnakumar,
 * Copyright (c) 2006      Chaitanya Patti,
 * Copyright (c) 2001-2006 Stony Brook University,
 * Copyright (c) 1997-2000 Columbia University
 *
 * For specific licensing information, see the COPYING file distributed with
 * this package, or get one from ftp://ftp.filesystems.org/pub/fist/COPYING.
 *
 * This Copyright notice must be kept intact and distributed with all
 * fistgen sources INCLUDING sources generated by fistgen.
 */
/*
 *  $Id: dentry.c,v 1.24 2006/12/13 17:00:21 kolya Exp $
 */



#ifdef HAVE_CONFIG_H
# include <config.h>
#endif /* HAVE_CONFIG_H */

#include "fist.h"
#include "raif.h"


/*
 * THIS IS A BOOLEAN FUNCTION: returns 1 if valid, 0 otherwise.
 */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
static int raif_d_revalidate(dentry_t *dentry, struct nameidata *nd)
#else
static int raif_d_revalidate(dentry_t *dentry, int flags)
#endif
{
	int err = 1; // default is valid (1); invalid is 0.
	int invalid = 0;
	dentry_t *hidden_dentry;
	int bindex, bnum;

	print_util_entry_location();


	/* 
	 * The revalidation must occur for all the common files across branches
	 */
	bnum = dbnum(dentry);
	for (bindex = 0; bindex < bnum; bindex++) {
		hidden_dentry  = dtohd_index(dentry, bindex);

		/* Lower level dentries could be NULL for certain indices
		 * Check for that. 
		 */ 
		if (!hidden_dentry || !hidden_dentry->d_op || !hidden_dentry->d_op->d_revalidate)
			continue;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
		if (dentry != dentry->d_parent)
			err = dentry->d_parent->d_op->d_revalidate(dentry->d_parent, nd);
#else
		if (dentry != dentry->d_parent)
			err = dentry->d_parent->d_op->d_revalidate(dentry->d_parent, flags);
#endif
		else // this is for both the cases of the macro
			err = 1; //the root is always valid ?

		if (!err)
			invalid = 1;
	}


	if (invalid)
		err = 0;
	//fist_print_dentry("revalidate out", dentry);
	print_util_exit_status(err);
	return err;
}

#if defined(RAIF_TAG_PAGE) || !defined(RAIF_TAG)
STATIC int raif_d_hash(dentry_t *dentry, qstr_t *name)
{
	int global_bindex = 0, err = 0;
	dentry_t *hidden_dentry;

	print_util_entry_location();

	hidden_dentry = dtohd_index(dentry, global_bindex);

	if (hidden_dentry && hidden_dentry->d_op && hidden_dentry->d_op->d_hash)
		hidden_dentry->d_op->d_hash(hidden_dentry, name);

	print_util_exit_status(err);
	return err;
}
#endif

STATIC int raif_d_compare(dentry_t *dentry, qstr_t *a, qstr_t *b)
{
	int err = 0;
#ifdef RAIF_DELAY_LOOKUP
#error "Not implemented"
#else
	dentry_t *hidden_dentry;

	print_util_entry_location();

	hidden_dentry = dtohd_index(dentry, 0);

	if (hidden_dentry->d_op && hidden_dentry->d_op->d_compare)
		err = hidden_dentry->d_op->d_compare(hidden_dentry, a, b);
	else
		err = ((a->len != b->len) || memcmp(a->name, b->name, b->len));
#endif
	print_util_exit_status(err);
	return err;
}


int raif_d_delete(dentry_t *dentry)
{
	dentry_t *hidden_dentry;
	int err = 0, global_err = 0;
	int bindex, bnum;

	print_entry_location();

	/* this could be a negative dentry, so check first */
	if (!dtopd(dentry)) {
		fist_dprint(6, "dentry without private data: %*s", dentry->d_name.len, dentry->d_name.name);
		goto out;
	}

	/* Delete all the lower dentries */
	bnum = dbnum(dentry);
	for (bindex = 0; bindex < bnum; bindex++) {
		/* hidden dentry could be NULL for certain indices */
		hidden_dentry = dtohd_index(dentry, bindex);

		// fist_print_dentry("D_DELETE IN", dentry);
		/* added b/c of changes to dput(): it calls d_drop on us */
		if (hidden_dentry &&
				hidden_dentry->d_op &&
				hidden_dentry->d_op->d_delete) {
			err = hidden_dentry->d_op->d_delete(hidden_dentry);
			if (err) {
				global_err = err;
			}
		}
	}

out:
	if (global_err) {
		err = global_err;
	}
	print_exit_status(err);
	return err;
}


void raif_d_release(dentry_t *dentry)
{
	dentry_t *hidden_dentry;
	int bindex, bnum;

	print_entry_location();
	//fist_print_dentry("raif_d_release IN dentry", dentry);

	/* this could be a negative dentry, so check first */
	if (!dtopd(dentry)) {
		fist_dprint(6, "dentry without private data: %*s", dentry->d_name.len, dentry->d_name.name);
		goto out;
	}

	/* Release all the hidden dentries */
	bnum = dbnum(dentry);
	for (bindex = 0; bindex < bnum; bindex++) {
		hidden_dentry = dtohd_index(dentry, bindex);
		/* decrement hidden dentry's counter and free its inode */
		if (hidden_dentry) {
			dput(hidden_dentry);
		}
		dtohd_index(dentry, bindex) = NULL;
	}

	/* free private data (raif_dentry_info) here */
	KFREE(dtohd_ptr(dentry));
	dtohd_ptr(dentry) = NULL;
	// XXX ARUN KFREE the rules here
	KFREE(dtopd(dentry));
	/* dtopd(dentry) */
	dentry->d_fsdata = NULL;	/* just to be safe */
out:
	print_exit_location();
}


/*
 * we don't really need raif_d_iput, because dentry_iput will call iput() if
 * raif_d_iput is not defined. We left this implemented for ease of
 * tracing/debugging.
 */
void raif_d_iput(dentry_t *dentry, inode_t *inode)
{
	print_entry_location();
	iput(inode);
	print_exit_location();
}


struct dentry_operations raif_dops = {
	d_revalidate:	raif_d_revalidate,
#if defined(RAIF_TAG_PAGE) || !defined(RAIF_TAG)
	d_hash:		raif_d_hash,
#endif
	d_compare:	raif_d_compare,
	d_release:	raif_d_release,
	d_delete:	raif_d_delete,
	d_iput:		raif_d_iput,
};


/*
 * Local variables:
 * c-basic-offset: 4
 * End:
 */
