/*
 * Copyright (c) 1997-2006 Erez Zadok
 * Copyright (c) 1997-2006 Nikolai Joukov
 * Copyright (c) 2001-2006 Stony Brook University
 * Copyright (c) 1997-2000 Columbia University
 *
 * For specific licensing information, see the COPYING file distributed with
 * this package, or get one from ftp://ftp.filesystems.org/pub/fist/COPYING.
 *
 * This Copyright notice must be kept intact and distributed with all
 * fistgen sources INCLUDING sources generated by fistgen.
 */
/*
 *  $Id: super.c,v 1.38 2006/12/31 01:43:58 crpatti Exp $
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif /* HAVE_CONFIG_H */
#include "fist.h"
#include "raif.h"

#ifdef RAIF_LB
extern struct proc_dir_entry *rlbfs_proc_root;
#endif /* RAIF_LB */

STATIC void raif_read_inode(inode_t *inode)
{
	print_entry_location();

	/* itopd(inode) */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19)
	inode->u.generic_ip = (struct raif_inode_info *) KMALLOC(sizeof(struct raif_inode_info), GFP_KERNEL);
#else
	inode->i_private = (struct raif_inode_info *) KMALLOC(sizeof(struct raif_inode_info), GFP_KERNEL);
#endif
	if (!itopd(inode)) {
		printk("<0>%s:%s:%d: No kernel memory!\n", __FILE__, __FUNCTION__, __LINE__);
		PASSERT(NULL);
	}
	init_itopd(inode);

	itohi_ptr(inode) = KMALLOC(sizeof(inode_t *) * sbnum(inode->i_sb), GFP_KERNEL);
	if (!itohi_ptr(inode)) {
		printk("<0>%s:%s:%d: No kernel memory!\n", __FILE__, __FUNCTION__, __LINE__);
		PASSERT(NULL);
	}
	
	init_itohi_ptr(inode, sbnum(inode->i_sb));

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	inode->i_version = ++event;	/* increment inode version */
#else
	inode->i_version++;
#endif

	itopd(inode)->uii_rdlock = SPIN_LOCK_UNLOCKED;
	itopd(inode)->uii_rdcount = 1;
	itopd(inode)->uii_hashsize = -1;
	INIT_LIST_HEAD(&itopd(inode)->uii_readdircache);

	inode->i_op = &raif_main_iops;
	inode->i_fop = &raif_main_fops;
	inode->i_mapping->a_ops = &raif_aops;

	print_exit_location();
}


STATIC void raif_put_inode(inode_t *inode)
{
	int bindex, bnum, write_flag;
	inode_t * hidden_inode = NULL;
	print_entry_location();
	fist_dprint(8, "%s i_count = %d, i_nlink = %d\n", __FUNCTION__,
		atomic_read(&inode->i_count), inode->i_nlink);
	/*
	 * This is really funky stuff:
	 * Basically, if i_count == 1, iput will then decrement it and this inode will be destroyed.
	 * It is currently holding a reference to the hidden inode.
	 * Therefore, it needs to release that reference by calling iput on the hidden inode.
	 * iput() _will_ do it for us (by calling our clear_inode), but _only_ if i_nlink == 0.
	 * The problem is, NFS keeps i_nlink == 1 for silly_rename'd files.
	 * So we must for our i_nlink to 0 here to trick iput() into calling our clear_inode.
	 */
	if (atomic_read(&inode->i_count) == 1) {
		/* If all the lower files have been deleted and this is 
		 * the last reference to this inode, we do not need to
		 * write out the inode before we release it.
		 */
		write_flag = 0;
		bnum = ibnum(inode);
		for (bindex = 0; bindex < bnum; bindex++) {
			hidden_inode = itohi_index(inode, bindex);
			if (hidden_inode == NULL)
				continue;
			if (hidden_inode->i_nlink > 0) {
				write_flag = 1;
				break;
			}
		}
		if (write_flag)
			write_inode_now(inode, 1);
		inode->i_nlink = 0;
	}

	print_exit_location();
}


/*
 * we now define delete_inode, because there are two VFS paths that may
 * destroy an inode: one of them calls clear inode before doing everything
 * else that's needed, and the other is fine.  This way we truncate the inode
 * size (and its pages) and then clear our own inode, which will do an iput
 * on our and the lower inode.
 */
STATIC void raif_delete_inode(inode_t *inode)
{
	print_entry_location();

	fist_checkinode(inode, "raif_delete_inode IN");
	inode->i_size = 0;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,14) 
	clear_inode(inode);
#else
	truncate_inode_pages(&inode->i_data, 0); 
	clear_inode(inode);
#endif

	print_exit_location();
}


/* final actions when unmounting a file system */
STATIC void raif_put_super(super_block_t *sb)
{
	int global_bindex, bnum;
	print_entry_location();

#ifdef RAIF_LB
	/* remove entry from /proc/rlbfs/ */
	/*remove_proc_entry(stopd(sb)->proc_name, rlbfs_proc_root);*/
	atomic_set(&stopd(sb)->die, 1);
	printk("waiting for the load balancing thread to die...\n");
	wait_event((stopd(sb)->clean), (atomic_read(&stopd(sb)->die) == 0));
#endif /* RAIF_LB */

	if (stopd(sb)) {
		bnum = sbnum(sb);
		for (global_bindex = 0; global_bindex < bnum; global_bindex++)
			mntput(stohm_index(sb, global_bindex));
		KFREE(stohs_ptr(sb));
		KFREE(stohiddenmnt_ptr(sb));
		KFREE(stopd(sb));
		stopd_lhs(sb) = NULL;
	}
	fist_dprint(6, "raif: released super\n");

	print_exit_location();
}


#ifdef NOT_NEEDED
/*
 * This is called in do_umount before put_super.
 * The superblock lock is not held yet.
 * We probably do not need to define this or call write_super
 * on the hidden_sb, because sync_supers() will get to hidden_sb
 * sooner or later.  But it is also called from file_fsync()...
 */
STATIC void raif_write_super(super_block_t *sb)
{
	return;
}
#endif /* NOT_NEEDED */

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
#define kstatfs statfs
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,18)
STATIC int raif_statfs(super_block_t *sb, struct kstatfs *buf)
#else
STATIC int raif_statfs(dentry_t *dentry, struct kstatfs *buf)
#endif
{
	int err = 0;
	super_block_t *hidden_sb;
	struct kstatfs rsb;
	int global_bindex, bindex1, bnum;
	unsigned short *hidden_s_dev;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18)
	super_block_t *sb = dentry->d_sb;
#endif
	print_entry_location();	 
	hidden_s_dev = (unsigned short *)KMALLOC(sizeof(unsigned short) * sbnum(sb), GFP_KERNEL);
	if (!hidden_s_dev) {
		err = -ENOMEM;
		goto out;
	}
	memset(buf, 0, sizeof(struct kstatfs));
	buf->f_type = RAIF_SUPER_MAGIC;
	
	bnum = sbnum(sb);
	for (global_bindex = 0; global_bindex < bnum; global_bindex++) {
		
		hidden_sb = stohs_index(sb, global_bindex);
	   
		/* store the current s_dev for checking any
		 * duplicate additions for super block data
		 */
		hidden_s_dev[global_bindex] = hidden_sb->s_dev;
		
		for (bindex1 = 0; bindex1 < global_bindex; bindex1++) {
			/* check the hidden super blocks with any previously 
			 * added super block data.  If match, continue
			 */
			if (hidden_s_dev[bindex1] == hidden_sb->s_dev) {
				continue;
			}
		}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,18)
		err = vfs_statfs(hidden_sb, &rsb);
#else
		err = vfs_statfs(dtohd_index(sb->s_root, global_bindex), &rsb);
#endif

		buf->f_bsize += rsb.f_bsize;
		buf->f_blocks += rsb.f_blocks;
		buf->f_bfree += rsb.f_bfree;
		buf->f_bavail += rsb.f_bavail;
		buf->f_files += rsb.f_files;
		buf->f_ffree += rsb.f_ffree;
	}

	KFREE(hidden_s_dev);
out:
	print_exit_status(err);
	return err;
}


/*
 * XXX: not implemented.  This is not allowed yet.
 * Should we call this on the hidden_sb?  Probably not.
 */
STATIC int raif_remount_fs(super_block_t *sb, int *flags, char *data)
{
	return -ENOSYS;
}


/*
 * Called by iput() when the inode reference count reached zero
 * and the inode is not hashed anywhere.  Used to clear anything
 * that needs to be, before the inode is completely destroyed and put
 * on the inode free list.
 */
STATIC void raif_clear_inode(inode_t *inode)
{
	int bindex, bnum;
	inode_t *hidden_inode;

	print_entry_location();

	//fist_checkinode(inode, "raif_clear_inode IN");
	/*
	 * Decrement a reference to a hidden_inode, which was incremented
	 * by our read_inode when it was created initially.
	 */
	bnum = ibnum(inode);
	for (bindex = 0; bindex < bnum; bindex++) {   
		hidden_inode = itohi_index(inode, bindex);
		if (hidden_inode == NULL)
			continue;
		iput(hidden_inode);
	}
	KFREE(itohi_ptr(inode));
	itohi_ptr(inode) = NULL;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	KFREE(itopd(inode));
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19)
	/* itopd(inode) */
	inode->u.generic_ip = NULL;
#else
	inode->i_private = NULL;
#endif
	print_exit_location();
}


/*
 * Called in do_umount() if the MNT_FORCE flag was used and this
 * function is defined.  See comment in linux/fs/super.c:do_umount().
 * Used only in nfs, to kill any pending RPC tasks, so that subsequent
 * code can actually succeed and won't leave tasks that need handling.
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,18)
STATIC void raif_umount_begin(super_block_t *sb)
#else
STATIC void raif_umount_begin(struct vfsmount *mnt, int flags)
#endif
{
	super_block_t *hidden_sb;
	int global_bindex, bnum;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18)
	super_block_t *sb = mnt->mnt_sb;
#endif

	print_entry_location();

	bnum = sbnum(sb);
	for (global_bindex = 0; global_bindex < bnum; global_bindex++) {
		hidden_sb = stohs_index(sb, global_bindex);
		if (hidden_sb && hidden_sb->s_op && hidden_sb->s_op->umount_begin)
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,18)
			hidden_sb->s_op->umount_begin(hidden_sb);
#else
			hidden_sb->s_op->umount_begin(stopd(sb)->usi_hidden_mnt[global_bindex], flags);
#endif
	}

	print_exit_location();
}


#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
/**
 * The entry given by dentry here is always a directory.
 * We can't just do dentry->d_parent, because it may
 * not be there, since this dentry could have been 
 * created by calling d_alloc_anon.
 */
static struct dentry *raif_get_parent(struct dentry *dentry) {
	return d_find_alias(dentry->d_inode);
}

struct export_operations raif_export_ops = {
	.get_parent	=	raif_get_parent
};
#endif

struct super_operations raif_sops =
{
	read_inode:	raif_read_inode,
	put_inode:	raif_put_inode,
	delete_inode:	raif_delete_inode,
	put_super:	raif_put_super,
	statfs:		raif_statfs,
	remount_fs:	raif_remount_fs,
	clear_inode:	raif_clear_inode,
	umount_begin:	raif_umount_begin,
};

/*
 * Local variables:
 * c-basic-offset: 4
 * End:
 */
